package com.jpattern.orm.generator.javassist;

import java.lang.reflect.Field;
import java.lang.reflect.Method;

import com.jpattern.javassist.CannotCompileException;
import com.jpattern.javassist.ClassPool;
import com.jpattern.javassist.CtClass;
import com.jpattern.javassist.CtNewMethod;
import com.jpattern.logger.ILogger;
import com.jpattern.orm.classmapper.IClassMapper;
import com.jpattern.orm.classmapper.IColumn;
import com.jpattern.orm.exception.OrmConfigurationException;
import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.generator.IOrmPersistor;
import com.jpattern.orm.generator.ResultSetMethodHelper;
import com.jpattern.orm.generator.wrapper.IWrapper;
import com.jpattern.orm.generator.wrapper.PrimitiveWrapper;
import com.jpattern.orm.generator.wrapper.VersionIncreaseHelper;
import com.jpattern.orm.logger.OrmLogger;

public class JavassistPersistorDump<T> {

	private static final String MAP_ROW_METHOD_SIGNATURE = "Object mapRow(String rowNamePrefix, java.sql.ResultSet rs, int rowNum)";
	private static final String PRIMARY_KEY_VALUES_METHOD_SIGNATURE = "Object[] primaryKeyValues(Object entity)";
	private static final String NOT_PRIMARY_KEY_VALUES_METHOD_SIGNATURE = "Object[] notPrimaryKeyValues(Object entity)";
	private static final String ALL_VALUES_METHOD_SIGNATURE = "Object[] allValues(Object entity)";
	private static final String ALL_NOT_GENERATED_VALUES_METHOD_SIGNATURE = "Object[] allNotGeneratedValues(Object entity)";
	private static final String UPDATE_GENERATED_VALUES_METHOD_SIGNATURE = "void updateGeneratedValues(java.sql.ResultSet rs, Object entity)";
	private static final String INCREASE_VERSION_METHOD_SIGNATURE = "void increaseVersion(Object entity, boolean firstVersionNumber)";
	private static final String PRIMARY_KEY_VALUES_AND_VERSION_METHOD_SIGNATURE = "Object[] primaryKeyAndVersionValues(Object entity)";
	private static final String START_METHOD = "{";
	private static final String END_METHOD = "}";
	private static final String METHOD_SCOPE = "public ";


	private final ILogger logger = OrmLogger.getOrmLogger(this.getClass());
	private final String createdClassName;
	private final IClassMapper<T> classMapper;

	public JavassistPersistorDump(String createdClassName, IClassMapper<T> classMapper) {
		this.createdClassName = createdClassName;
		this.classMapper = classMapper;
	}


	public Class<IOrmPersistor<T>> createClass() throws CannotCompileException, SecurityException, OrmConfigurationException, NoSuchMethodException {

		final ClassPool pool = ClassPool.getDefault();

		final CtClass evalClass = pool.makeClass(createdClassName);
		evalClass.setInterfaces( new CtClass[] { pool.makeClass(IOrmPersistor.class.getCanonicalName()) });

		createAllValuesMethod(evalClass);
		createAllNotGeneratedValuesMethod(evalClass);
		createNotPrimaryKeyValuesMethod(evalClass);
		createPrimaryKeyValuesMethod(evalClass);
		createPrimaryKeyAndVersionValuesMethod(evalClass);
		createIncreaseVersionMethod(evalClass);
		createMapRowMethod(evalClass);
		createUpdateGeneratedValuesMethod(evalClass);

		@SuppressWarnings("unchecked")
		final
		Class<IOrmPersistor<T>> clazz = evalClass.toClass();
		return clazz;
	}

	private void createUpdateGeneratedValuesMethod(CtClass evalClass) throws SecurityException, OrmConfigurationException, NoSuchMethodException, CannotCompileException {
		final StringBuffer methodBody = new StringBuffer();
		final String className = classMapper.getMappedClass().getCanonicalName();
		final String[] allColumnNames = classMapper.getAllGeneratedColumnJavaNames();

		methodBody.append("try {\n");
		methodBody.append( className + " realEntity = " + "(" + className + ") entity; \n" );
		int i = 1;
		for (final String columnJavaName : allColumnNames) {
			final IColumn column = classMapper.getColumnWithJavaName(columnJavaName);
			final Field field = column.getField();
			final Method resultSetMethod = ResultSetMethodHelper.findResultSetGetterForInt(field.getType());

			if ( ResultSetMethodHelper.isWrappedType(field.getType()) ) {
				final IWrapper<?, ?> wrapper = ResultSetMethodHelper.getWrapper(field.getType());
				methodBody.append("realEntity." + column.getSetter().getName() + "(" + wrapper.getClass().getCanonicalName() + ".staticWrap( rs." + resultSetMethod.getName() + "(" + i + ") ) );\n" );
			}
			else if (resultSetMethod.getReturnType().isPrimitive() && !field.getType().isPrimitive()) {
				methodBody.append("realEntity." + column.getSetter().getName() + "(" + PrimitiveWrapper.class.getCanonicalName() + ".wrap( rs." + resultSetMethod.getName() + "(" + i + ") ) );\n" );
			}
			else {
				methodBody.append("realEntity." + column.getSetter().getName() + "(" + "rs." + resultSetMethod.getName() + "(" + i + ") );\n" );
			}

			i++;
		}

		methodBody.append("} catch ( Exception e) { throw new " + OrmException.class.getCanonicalName() + "(e); }");
		createMethod(evalClass, UPDATE_GENERATED_VALUES_METHOD_SIGNATURE, methodBody.toString());
	}


	private void createMapRowMethod(CtClass evalClass) throws CannotCompileException, SecurityException, OrmConfigurationException, NoSuchMethodException {
		final StringBuffer methodBody = new StringBuffer();
		final String className = classMapper.getMappedClass().getCanonicalName();
		final String[] allColumnNames = classMapper.getAllColumnJavaNames();

		methodBody.append(className + " bean = new " + className + "();");
		methodBody.append("try {");
		for (final String columnJavaName : allColumnNames) {
			final IColumn column = classMapper.getColumnWithJavaName(columnJavaName);
			final Field field = column.getField();
			final Method resultSetMethod = ResultSetMethodHelper.findResultSetGetterForString(field.getType());
			methodBody.append("bean." + column.getSetter().getName() + "(" );

			if ( ResultSetMethodHelper.isWrappedType(field.getType()) ) {
				final IWrapper<?, ?> wrapper = ResultSetMethodHelper.getWrapper(field.getType());
				methodBody.append( wrapper.getClass().getCanonicalName() + ".staticWrap(rs." + resultSetMethod.getName() + "( rowNamePrefix + \"" + column.getName() + "\") ) "  );
			}
			else
				if (resultSetMethod.getReturnType().isPrimitive() && !field.getType().isPrimitive()) {
					methodBody.append( PrimitiveWrapper.class.getCanonicalName() + ".wrap(rs." + resultSetMethod.getName() + "( rowNamePrefix + \"" + column.getName() + "\") ) "  );
				}
				else {
					methodBody.append("rs." + resultSetMethod.getName() + "( rowNamePrefix + \"" + column.getName() + "\") "  );
				}
			methodBody.append( " );" );
		}

		methodBody.append("} catch ( Exception e) { throw new " + OrmException.class.getCanonicalName() + "(e); }");
		methodBody.append("return bean;");
		createMethod(evalClass, MAP_ROW_METHOD_SIGNATURE, methodBody.toString());
	}


	private void createAllValuesMethod(CtClass evalClass) throws CannotCompileException, SecurityException, OrmConfigurationException, NoSuchMethodException {
		final String[] javaColumnNames = classMapper.getAllColumnJavaNames();
		final StringBuffer methodBody = new StringBuffer();
		buildValuesMethod(javaColumnNames, methodBody);
		createMethod(evalClass, ALL_VALUES_METHOD_SIGNATURE, methodBody.toString());
	}

	private void createAllNotGeneratedValuesMethod(CtClass evalClass) throws CannotCompileException, SecurityException, OrmConfigurationException, NoSuchMethodException {
		final String[] javaColumnNames = classMapper.getAllNotGeneratedColumnJavaNames();
		final StringBuffer methodBody = new StringBuffer();
		buildValuesMethod(javaColumnNames, methodBody);
		createMethod(evalClass, ALL_NOT_GENERATED_VALUES_METHOD_SIGNATURE, methodBody.toString());
	}

	private void createNotPrimaryKeyValuesMethod(CtClass evalClass) throws CannotCompileException, SecurityException, OrmConfigurationException, NoSuchMethodException {
		final String[] javaColumnNames = classMapper.getNotPrimaryKeyColumnJavaNames();
		final StringBuffer methodBody = new StringBuffer();
		buildValuesMethod(javaColumnNames, methodBody);
		createMethod(evalClass, NOT_PRIMARY_KEY_VALUES_METHOD_SIGNATURE, methodBody.toString());
	}

	private void createPrimaryKeyValuesMethod(CtClass evalClass) throws CannotCompileException, SecurityException, OrmConfigurationException, NoSuchMethodException {
		final String[] javaColumnNames = classMapper.getPrimaryKeyColumnJavaNames();
		final StringBuffer methodBody = new StringBuffer();
		buildValuesMethod(javaColumnNames, methodBody);
		createMethod(evalClass, PRIMARY_KEY_VALUES_METHOD_SIGNATURE, methodBody.toString());
	}

	private void createPrimaryKeyAndVersionValuesMethod(CtClass evalClass) throws CannotCompileException, SecurityException, OrmConfigurationException, NoSuchMethodException {
		final String[] javaColumnNames = classMapper.getPrimaryKeyAndVersionColumnJavaNames();
		final StringBuffer methodBody = new StringBuffer();
		buildValuesMethod(javaColumnNames, methodBody);
		createMethod(evalClass, PRIMARY_KEY_VALUES_AND_VERSION_METHOD_SIGNATURE, methodBody.toString());
	}

	private void createIncreaseVersionMethod(CtClass evalClass) throws CannotCompileException {
		final StringBuffer methodBody = new StringBuffer();

		if (classMapper.getTableMap().isVersionable()) {
			final String className = classMapper.getMappedClass().getCanonicalName();
			final IColumn column = classMapper.getColumnWithJavaName( classMapper.getTableMap().getVersionJavaFieldName() );
			final Method getter = column.getGetter();
			final Method setter = column.getSetter();
			methodBody.append( className + " realEntity = " + "(" + className + ") entity; " );
			methodBody.append( "realEntity." + setter.getName() + "( "+ VersionIncreaseHelper.class.getCanonicalName() + ".increase( realEntity." + getter.getName() + "() , firstVersionNumber) );" );
		}

		createMethod(evalClass, INCREASE_VERSION_METHOD_SIGNATURE, methodBody.toString());
	}

	private void createMethod(CtClass evalClass, String methodSignature, String methodBody) throws CannotCompileException {
		final String fullMethod = "\n" + METHOD_SCOPE + methodSignature + START_METHOD + "\n" + methodBody + "\n" + END_METHOD;
		logger.debug("createMethod", "Generated persistor for Class [" + classMapper.getMappedClass() + "] - Creating method with signature [" + METHOD_SCOPE + methodSignature + "]");
		logger.trace("createMethod", "Generated method code:");
		logger.trace("createMethod", fullMethod);
		evalClass.addMethod(CtNewMethod.make(fullMethod, evalClass));
	}

	private void buildValuesMethod(String[] javaColumnNames, StringBuffer methodBody) throws SecurityException, OrmConfigurationException, NoSuchMethodException {
		final String className = classMapper.getMappedClass().getCanonicalName();
		methodBody.append( className + " realEntity = " + "(" + className + ") entity; " );
		methodBody.append( "int i = 0;"  );
		methodBody.append( "Object[] result = new Object[" + javaColumnNames.length + "];"  );
		for (final String javaColumnName : javaColumnNames) {
			final IColumn column = classMapper.getColumnWithJavaName(javaColumnName);
			final Method getter = column.getGetter();

			if ( ResultSetMethodHelper.isWrappedType(column.getField().getType()) ) {
				final IWrapper<?, ?> wrapper = ResultSetMethodHelper.getWrapper(column.getField().getType());
				methodBody.append(  "result[i++] = " + wrapper.getClass().getCanonicalName() + ".staticUnWrap(realEntity." + getter.getName() + "());");			}
			else
				if (column.getField().getType().isPrimitive()) {
					methodBody.append(  "result[i++] = " + PrimitiveWrapper.class.getCanonicalName() + ".wrap(realEntity." + getter.getName() + "());");
				} else {
					methodBody.append(  "result[i++] = realEntity." + getter.getName() + "();");
				}
		}
		methodBody.append( "return result;" );
	}

}
